@@ -317,8 +317,12 @@ def _process_configuration(section):
                      'getint', None)
     _process_setting(section, 'agent_limits.slow_sql_stack_trace',
                      'getint', None)
+    _process_setting(section, 'agent_limits.max_sql_connections',
+                     'getint', None)
     _process_setting(section, 'agent_limits.sql_explain_plans',
                      'getint', None)
+    _process_setting(section, 'agent_limits.sql_explain_plans_per_harvest',
+                     'getint', None)
     _process_setting(section, 'agent_limits.slow_sql_data',
                      'getint', None)
     _process_setting(section, 'agent_limits.merge_stats_maximum',
@@ -361,6 +365,8 @@ def _process_configuration(section):
                      'getboolean', None)
     _process_setting(section, 'debug.log_agent_initialization',
                      'getboolean', None)
+    _process_setting(section, 'debug.log_explain_plan_queries',
+                     'getboolean', None)
     _process_setting(section, 'debug.enable_coroutine_profiling',
                      'getboolean', None)
     _process_setting(section, 'debug.record_transaction_failure',
