@@ -28,6 +28,8 @@ from newrelic.core.internal_metrics import (InternalTrace,
 from newrelic.core.xray_session import XraySession
 from newrelic.core.profile_sessions import profile_session_manager
 
+from .database_utils import SQLConnections
+
 _logger = logging.getLogger(__name__)
 
 class Application(object):
@@ -706,8 +708,10 @@ class Application(object):
                 if settings.debug.record_transaction_failure:
                     raise
 
-            if profile_samples and data.path in \
-                    self._stats_engine.xray_sessions:
+            if (profile_samples and (data.path in
+                    self._stats_engine.xray_sessions or
+                    'WebTransaction/Agent/__profiler__' in
+                    self._stats_engine.xray_sessions)):
 
                 try:
                     background_task, samples = profile_samples
@@ -716,8 +720,16 @@ class Application(object):
                             'stack_traces[sample]', len(samples))
 
                     tr_type = 'BACKGROUND' if background_task else 'REQUEST'
-                    self.profile_manager.add_stack_traces(self._app_name,
-                            data.path, tr_type, samples)
+
+                    if data.path in self._stats_engine.xray_sessions:
+                        self.profile_manager.add_stack_traces(self._app_name,
+                                data.path, tr_type, samples)
+
+                    if ('WebTransaction/Agent/__profiler__' in
+                            self._stats_engine.xray_sessions):
+                        self.profile_manager.add_stack_traces(self._app_name,
+                                'WebTransaction/Agent/__profiler__', tr_type,
+                                samples)
 
                 except Exception:
                     _logger.exception('Building xray profile tree has failed.'
@@ -1284,23 +1296,33 @@ class Application(object):
                             self._active_session.send_errors(error_data)
 
                     if configuration.collect_traces:
-                        slow_sql_data = stats.slow_sql_data()
+                        connections = SQLConnections(
+                                configuration.agent_limits.max_sql_connections)
 
-                        internal_metric('Supportability/Harvest/Counts/'
-                                'sql_trace_data', len(slow_sql_data))
+                        with connections:
+                            if configuration.slow_sql.enabled:
+                                slow_sql_data = stats.slow_sql_data(
+                                        connections)
 
-                        if slow_sql_data:
-                            self._active_session.send_sql_traces(slow_sql_data)
+                                internal_metric('Supportability/Harvest/'
+                                        'Counts/sql_trace_data',
+                                        len(slow_sql_data))
 
-                        slow_transaction_data = stats.transaction_trace_data()
+                                if slow_sql_data:
+                                    self._active_session.send_sql_traces(
+                                            slow_sql_data)
 
-                        internal_metric('Supportability/Harvest/Counts/'
-                                'transaction_sample_data',
-                                len(slow_transaction_data))
+                            slow_transaction_data = (
+                                    stats.transaction_trace_data(
+                                    connections))
 
-                        if slow_transaction_data:
-                            self._active_session.send_transaction_traces(
-                                    slow_transaction_data)
+                            internal_metric('Supportability/Harvest/Counts/'
+                                    'transaction_sample_data',
+                                    len(slow_transaction_data))
+
+                            if slow_transaction_data:
+                                self._active_session.send_transaction_traces(
+                                        slow_transaction_data)
 
                     # Fetch agent commands sent from the data collector
                     # and process them.
