@@ -218,7 +218,9 @@ _settings.agent_limits.data_collector_timeout = 30.0
 _settings.agent_limits.transaction_traces_nodes = 2000
 _settings.agent_limits.sql_query_length_maximum = 16384
 _settings.agent_limits.slow_sql_stack_trace = 30
+_settings.agent_limits.max_sql_connections = 4
 _settings.agent_limits.sql_explain_plans = 30
+_settings.agent_limits.sql_explain_plans_per_harvest = 60
 _settings.agent_limits.slow_sql_data = 10
 _settings.agent_limits.merge_stats_maximum = 5
 _settings.agent_limits.errors_per_transaction = 5
@@ -245,6 +247,7 @@ _settings.debug.log_thread_profile_payload = False
 _settings.debug.log_normalization_rules = False
 _settings.debug.log_raw_metric_data = False
 _settings.debug.log_normalized_metric_data = False
+_settings.debug.log_explain_plan_queries = False
 _settings.debug.record_transaction_failure = False
 _settings.debug.enable_coroutine_profiling = False
 _settings.debug.explain_plan_obfuscation = 'simple'
