@@ -11,17 +11,7 @@ import weakref
 from newrelic.core.internal_metrics import (internal_trace, internal_metric)
 from newrelic.core.config import global_settings
 
-# Various actions are dependent on the database being used. For most
-# flexibility we use the name of the Python module used as client.
-# falling back to allowing a string name also to be passed in. This does
-# mean we are limited to what specific client modules we are supporting
-# and not easy for a user to apply generic DBAPI2 instrumentation to
-# another module. Need to investigate a better way of handling that at
-# some point.
-
-def _dbapi2_module_name(dbapi2_module):
-    return (dbapi2_module and hasattr(dbapi2_module, '__name__') and
-            dbapi2_module.__name__ or dbapi2_module)
+_logger = logging.getLogger(__name__)
 
 # Obfuscation of SQL is done when reporting SQL statements back to the
 # data collector so that sensitive information is not being passed.
@@ -58,26 +48,16 @@ _any_quotes_re = re.compile(_any_quotes_p)
 _int_re = re.compile(r'(?<!:)\b\d+\b')
 
 _quotes_table = {
-    'MySQLdb': _any_quotes_re,
-    'pymysql': _any_quotes_re,
-    'oursql': _any_quotes_re,
+    'single': _single_quotes_re,
+    'double': _double_quotes_re,
+    'single+double': _any_quotes_re,
 }
 
 _quotes_default = _single_quotes_re
 
 @internal_trace('Supportability/DatabaseUtils/Calls/obfuscate_sql')
-def _obfuscate_sql(sql, dbapi2_module=None):
-    quoting_style = getattr(dbapi2_module, '_nr_quoting_style', None)
-
-    if quoting_style == 'single':
-        quotes_re = _single_quotes_re
-    elif quoting_style == 'double':
-        quotes_re = _double_quotes_re
-    elif quoting_style == 'single+double':
-        quotes_re = _any_quotes_re
-    else:
-        name = _dbapi2_module_name(dbapi2_module)
-        quotes_re = _quotes_table.get(name, _quotes_default)
+def _obfuscate_sql(sql, database):
+    quotes_re = _quotes_table.get(database.quoting_style, _single_quotes_re)
 
     # Substitute quoted strings first.
 
@@ -128,7 +108,7 @@ _normalize_whitespace_3_p = r'(?<!\w)\s+'
 _normalize_whitespace_3_re = re.compile(_normalize_whitespace_3_p)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/normalize_sql')
-def _normalize_sql(sql, dbapi2_module):
+def _normalize_sql(sql):
     # Note we that do this as a series of regular expressions as
     # using '|' in regular expressions is more expensive.
 
@@ -167,7 +147,7 @@ def _normalize_sql(sql, dbapi2_module):
 # might be preceded or followed by punctuation which we can expect in
 # context of SQL statements.
 #
-# TODO This is always going to do what we require, but it is only used
+# TODO This isn't always going to do what we require, but it is only used
 # now for cases below which never get invoked, so is okay for now.
 
 _identifier_re = re.compile('[\',"`\[\]\(\)]*')
@@ -180,7 +160,7 @@ def _extract_identifier(token):
 _uncomment_sql_p = r'/\*.*?\*/'
 _uncomment_sql_re = re.compile(_uncomment_sql_p, re.DOTALL)
 
-def _uncomment_sql(sql, dbapi2_module):
+def _uncomment_sql(sql):
     return _uncomment_sql_re.sub('', sql)
 
 # Parser routines for the different SQL statement operation types.
@@ -248,12 +228,12 @@ _parse_from_p = '\s+FROM\s+' + _parse_identifier_p
 _parse_from_re = re.compile(_parse_from_p, re.IGNORECASE)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_select')
-def _parse_select(sql, dbapi2_module):
+def _parse_select(sql):
     m = _parse_from_re.search(sql)
     return m and next(s for s in m.groups()[1:] if s).lower() or ''
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_delete')
-def _parse_delete(sql, dbapi2_module):
+def _parse_delete(sql):
     m = _parse_from_re.search(sql)
     return m and next(s for s in m.groups()[1:] if s).lower() or ''
 
@@ -261,7 +241,7 @@ _parse_into_p = '\s+INTO\s+' + _parse_identifier_p
 _parse_into_re = re.compile(_parse_into_p, re.IGNORECASE)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_insert')
-def _parse_insert(sql, dbapi2_module):
+def _parse_insert(sql):
     m = _parse_into_re.search(sql)
     return m and next(s for s in m.groups()[1:] if s).lower() or ''
 
@@ -269,7 +249,7 @@ _parse_update_p = '\s*UPDATE\s+' + _parse_identifier_p
 _parse_update_re = re.compile(_parse_update_p, re.IGNORECASE)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_update')
-def _parse_update(sql, dbapi2_module):
+def _parse_update(sql):
     m = _parse_update_re.search(sql)
     return m and next(s for s in m.groups()[1:] if s).lower() or ''
 
@@ -277,12 +257,12 @@ _parse_table_p = '\s+TABLE\s+' + _parse_identifier_p
 _parse_table_re = re.compile(_parse_table_p, re.IGNORECASE)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_create')
-def _parse_create(sql, dbapi2_module):
+def _parse_create(sql):
     m = _parse_table_re.search(sql)
     return m and next(s for s in m.groups()[1:] if s).lower() or ''
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_drop')
-def _parse_drop(sql, dbapi2_module):
+def _parse_drop(sql):
     m = _parse_table_re.search(sql)
     return m and next(s for s in m.groups()[1:] if s).lower() or ''
 
@@ -293,42 +273,42 @@ _parse_call_p = r'\s*CALL\s+(?!\()(\w+)'
 _parse_call_re = re.compile(_parse_call_p, re.IGNORECASE)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_call')
-def _parse_call(sql, dbapi2_module):
+def _parse_call(sql):
     return _parse_default(sql, _parse_call_re)
 
 _parse_show_p = r'\s*SHOW\s+(.*)'
 _parse_show_re = re.compile(_parse_show_p, re.IGNORECASE | re.DOTALL)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_show')
-def _parse_show(sql, dbapi2_module):
+def _parse_show(sql):
     return _parse_default(sql, _parse_show_re)
 
 _parse_set_p = r'\s*SET\s+(.*?)\W+.*'
 _parse_set_re = re.compile(_parse_set_p, re.IGNORECASE | re.DOTALL)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_set')
-def _parse_set(sql, dbapi2_module):
+def _parse_set(sql):
     return _parse_default(sql, _parse_set_re)
 
 _parse_exec_p = r'\s*EXEC\s+(?!\()(\w+)'
 _parse_exec_re = re.compile(_parse_exec_p, re.IGNORECASE)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_exec')
-def _parse_exec(sql, dbapi2_module):
+def _parse_exec(sql):
     return _parse_default(sql, _parse_exec_re)
 
 _parse_execute_p = r'\s*EXECUTE\s+(?!\()(\w+)'
 _parse_execute_re = re.compile(_parse_execute_p, re.IGNORECASE)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_execute')
-def _parse_execute(sql, dbapi2_module):
+def _parse_execute(sql):
     return _parse_default(sql, _parse_execute_re)
 
 _parse_alter_p = r'\s*ALTER\s+(?!\()(\w+)'
 _parse_alter_re = re.compile(_parse_alter_p, re.IGNORECASE)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target_alter')
-def _parse_alter(sql, dbapi2_module):
+def _parse_alter(sql):
     return _parse_default(sql, _parse_alter_re)
 
 _operation_table = {
@@ -350,15 +330,15 @@ _parse_operation_p = r'(\w+)'
 _parse_operation_re = re.compile(_parse_operation_p)
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_operation')
-def _parse_operation(sql, dbapi2_module):
+def _parse_operation(sql):
     match = _parse_operation_re.search(sql)
     operation = match and match.group(1).lower() or ''
     return operation if operation in _operation_table else ''
 
 @internal_trace('Supportability/DatabaseUtils/Calls/parse_target')
-def _parse_target(sql, dbapi2_module, operation):
+def _parse_target(sql, operation):
     parse = _operation_table.get(operation, None)
-    return parse and parse(sql, dbapi2_module) or ''
+    return parse and parse(sql) or ''
 
 # For explain plan obfuscation, the regular expression for matching the
 # explain plan needs to give precedence to replacing double quotes from
@@ -445,115 +425,292 @@ def _obfuscate_explain_plan_postgresql(columns, rows, mask=None):
 
     return columns, rows
 
-_explain_plan_table = {
-    'MySQLdb': ('EXPLAIN', ('select',), None),
-    'ibm_db_dbi': ('EXPLAIN',
-            ('select', 'insert', 'update', 'delete'), None),
-    'oursql': ('EXPLAIN', ('select',), None),
-    'pymysql': ('EXPLAIN', ('select',), None),
-    'postgresql.interface.proboscis.dbapi2': ('EXPLAIN',
-            ('select', 'insert', 'update', 'delete'),
-            _obfuscate_explain_plan_postgresql),
-    'psycopg2': ('EXPLAIN',
-            ('select', 'insert', 'update', 'delete'),
-            _obfuscate_explain_plan_postgresql),
-    'psycopg2ct': ('EXPLAIN',
-            ('select', 'insert', 'update', 'delete'),
-            _obfuscate_explain_plan_postgresql),
-    'sqlite3.dbapi2': ('EXPLAIN QUERY PLAN',
-            ('select', 'insert', 'update', 'delete')),
+_obfuscate_explain_plan_table = {
+    'PostgreSQL': _obfuscate_explain_plan_postgresql
 }
 
-@internal_trace('Supportability/DatabaseUtils/Calls/explain_plan')
-def _explain_plan(sql, dbapi2_module, connect_params, cursor_params,
-        execute_params, format):
+def _obfuscate_explain_plan(database, columns, rows):
+    obfuscator = _obfuscate_explain_plan_table.get(database.name)
+    if obfuscator:
+        return obfuscator(columns, rows)
+    return columns, rows
 
-    if dbapi2_module is None:
-        return None
+class SQLConnection(object):
 
-    name = _dbapi2_module_name(dbapi2_module)
+    def __init__(self, database, connection):
+        self.database = database
+        self.connection = connection
+        self.cursors = {}
 
-    if name is None:
-        return None
+    def cursor(self, args=(), kwargs={}):
+        key = (args, frozenset(kwargs.items()))
 
-    if connect_params is None:
-        return None
+        cursor = self.cursors.get(key)
 
-    query = None
-    obfuscator = None
+        if cursor is None:
+            cursor = self.connection.cursor(*args, **kwargs)
+            self.cursors[key] = cursor
 
-    command = getattr(dbapi2_module, '_nr_explain_query', None)
+        return cursor
 
-    if command:
-        operations = dbapi2_module._nr_explain_stmts
-        if dbapi2_module._nr_database_name == 'PostgreSQL':
-            obfuscator = _obfuscate_explain_plan_postgresql
-    else:
-        command, operations, obfuscator = _explain_plan_table.get(
-            name, (None, None, None))
+    def cleanup(self):
+        _logger.debug('Cleanup database connection for %r.',
+                self.database)
 
-    if not command:
-        return None
-
-    # Don't do an explain plan for operations which the database is
-    # believed not to support doing so. Also restrict them to the basic
-    # SQL operation types even though some database such as PostgreSQL
-    # can support others. This is to avoid errors being logged in
-    # database log file when explain plan done on the wrong thing. Not
-    # going through the SQL cache here as only doing this for a small
-    # number of SQL statements per application per harvest cycle.
-
-    operation = _parse_operation(sql, dbapi2_module)
-
-    if operation not in operations:
-        return None
-
-    query = '%s %s' % (command, sql)
-
-    args, kwargs = connect_params
-    try:
-        connection = dbapi2_module.connect(*args, **kwargs)
         try:
-            if cursor_params is not None:
-                args, kwargs = cursor_params
-                cursor = connection.cursor(*args, **kwargs)
-            else:
-                cursor = connection.cursor()
-            try:
-                if execute_params is not None:
-                    cursor.execute(query, execute_params)
-                else:
-                    cursor.execute(query)
-                columns = []
-                if cursor.description:
-                    for column in cursor.description:
-                        columns.append(column[0])
-                rows = cursor.fetchall()
-                if not columns and not rows:
-                    return None
-                if obfuscator and format != 'raw':
-                    columns, rows = obfuscator(columns, rows)
-                return (columns, rows)
-            except Exception:
-                pass
-            finally:
-                cursor.close()
-        finally:
-            try:
-                connection.rollback()
-            except (AttributeError, dbapi2_module.NotSupportedError):
-                pass
-            connection.close()
+            self.connection.rollback()
+            pass
+        except (AttributeError, self.database.NotSupportedError):
+            pass
+
+        self.connection.close()
+
+class SQLConnections(object):
+
+    def __init__(self, maximum=4):
+        self.connections = []
+        self.maximum = maximum
+
+        _logger.debug('Creating SQL connections cache %r.', self)
+
+    def connection(self, database, args, kwargs):
+        key = (database.client, args, kwargs)
+
+        connection = None
+
+        for i, item in enumerate(self.connections):
+            if item[0] == key:
+                connection = item[1]
+
+                # Move to back of list so we know which is the
+                # most recently used all the time.
+
+                item = self.connections.pop(i)
+                self.connections.append(item)
+
+                break
+
+        if connection is None:
+            # If we are at the maximum number of connections to
+            # keep hold of, pop the one which has been used the
+            # longest amount of time.
+
+            if len(self.connections) == self.maximum:
+                connection = self.connections.pop(0)[1]
+
+                internal_metric('Supportability/DatabaseUtils/Counts/'
+                                'drop_database_connection', 1)
+
+                _logger.debug('Drop database connection for %r as '
+                        'reached maximum of %r.', connection.database.client,
+                        self.maximum)
+
+                connection.cleanup()
+
+            connection = SQLConnection(database,
+                    database.connect(*args, **kwargs))
+
+            self.connections.append((key, connection))
+
+            internal_metric('Supportability/DatabaseUtils/Counts/'
+                            'create_database_connection', 1)
+
+            _logger.debug('Created database connection for %r.',
+                    database.client)
+
+        return connection
+
+    def cleanup(self):
+        _logger.debug('Cleaning up SQL connections cache %r.', self)
+
+        for key, connection in self.connections:
+            connection.cleanup()
+
+        self.connections = []
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc, value, tb):
+        self.cleanup()
+
+@internal_trace('Supportability/DatabaseUtils/Calls/explain_plan')
+def _explain_plan(connections, sql, database, connect_params, cursor_params,
+        sql_parameters, execute_params):
+
+    query = '%s %s' % (database.explain_query, sql)
+
+    settings = global_settings()
+
+    if settings.debug.log_explain_plan_queries:
+        _logger.debug('Executing explain plan for %r on %r.', query,
+                database.client)
+
+    try:
+        args, kwargs = connect_params
+        connection = connections.connection(database, args, kwargs)
+
+        if cursor_params is not None:
+            args, kwargs = cursor_params
+            cursor = connection.cursor(args, kwargs)
+        else:
+            cursor = connection.cursor()
+
+        if execute_params is not None:
+            args, kwargs = execute_params
+        else:
+            args, kwargs = ((), {})
+
+        # If sql_parameters is None them args would need
+        # to be an empty sequence. Don't pass it just in
+        # case it wasn't for some reason, and only supply
+        # kwargs. Right now the only time we believe that
+        # passing in further params is needed is with
+        # oursql cursor execute() method, which has
+        # proprietary arguments outside of the DBAPI2
+        # specification.
+
+        if sql_parameters is not None:
+            cursor.execute(query, sql_parameters, *args, **kwargs)
+        else:
+            cursor.execute(query, **kwargs)
+
+        columns = []
+
+        if cursor.description:
+            for column in cursor.description:
+                columns.append(column[0])
+
+        rows = cursor.fetchall()
+
+        if not columns and not rows:
+            return None
+
+        return (columns, rows)
+
     except Exception:
-        pass
+        if settings.debug.log_explain_plan_queries:
+            _logger.exception('Error occurred when executing explain '
+                    'plan for %r on %r where cursor_params=%r and '
+                    'execute_params=%r.', query, database.client,
+                    cursor_params, execute_params)
 
     return None
 
+def explain_plan(connections, sql_statement, connect_params, cursor_params,
+        sql_parameters, execute_params, sql_format):
+
+    # If no parameters supplied for creating database connection
+    # then mustn't have been a candidate for explain plans in the
+    # first place, so skip it.
+
+    if connect_params is None:
+        return
+
+    # Determine if we even know how to perform explain plans for
+    # this particular database.
+
+    database = sql_statement.database
+
+    if sql_statement.operation not in database.explain_stmts:
+        return
+
+    details = _explain_plan(connections, sql_statement.sql, database,
+            connect_params, cursor_params, sql_parameters, execute_params)
+
+    if details is not None and sql_format != 'raw':
+        return _obfuscate_explain_plan(database, *details)
+
+    return details
+
+# Wrapper for information about a specific database. We haven't yet
+# added high level instrumentation modules for all databases we know
+# of and so we still maintain a table here indexed by Python module
+# name for some.
+
+DATABASE_MODULES = {
+    'ibm_db_dbi': 'DB2'
+}
+
+DATABASE_DEFINTIONS = {
+    'DB2': {
+        'quoting_style': 'single',
+        'explain_query': 'EXPLAIN',
+        'explain_stmts': ('select', 'insert', 'update', 'delete')
+    }
+}
+
+class SQLDatabase(object):
+
+    def __init__(self, dbapi2_module):
+        self.dbapi2_module = dbapi2_module
+
+    def __getattr__(self, name):
+        return getattr(self.dbapi2_module, name)
+
+    @property
+    def name(self):
+        name = getattr(self.dbapi2_module, '_nr_database_name', None)
+
+        if name is None:
+            name = getattr(self.dbapi2_module, None)
+
+            if name:
+                name = DATABASE_MODULES.get(name)
+
+        return name
+
+    @property
+    def client(self):
+        name = getattr(self.dbapi2_module, '__name__', None)
+        if name is None:
+            name = getattr(self.dbapi2_module, '__file__', None)
+        if name is None:
+            name = str(self.dbapi2_module)
+        return name
+
+    @property
+    def quoting_style(self):
+        result = getattr(self.dbapi2_module, '_nr_quoting_style', None)
+
+        if result is None:
+            details = DATABASE_DEFINTIONS.get(self.name)
+            if details:
+                result = details['quoting_style']
+            else:
+                result = 'single'
+
+        return result
+
+    @property
+    def explain_query(self):
+        result = getattr(self.dbapi2_module, '_nr_explain_query', None)
+
+        if result is None:
+            details = DATABASE_DEFINTIONS.get(self.name)
+            if details:
+                result = details['explain_query']
+
+        return result
+
+    @property
+    def explain_stmts(self):
+        result = getattr(self.dbapi2_module, '_nr_explain_stmts', None)
+
+        if result is None:
+            details = DATABASE_DEFINTIONS.get(self.name)
+            if details:
+                result = details['explain_stmts']
+            else:
+                result = ()
+
+        return result
+
 class SQLStatement(object):
 
-    def __init__(self, sql, dbapi2_module=None):
+    def __init__(self, sql, database=None):
         self.sql = sql
-        self.dbapi2_module = dbapi2_module
+        self.database = database
 
         self._operation = None
         self._target = None
@@ -565,35 +722,31 @@ class SQLStatement(object):
     @property
     def operation(self):
         if self._operation is None:
-            self._operation = _parse_operation(self.uncommented,
-                    self.dbapi2_module)
+            self._operation = _parse_operation(self.uncommented)
         return self._operation
 
     @property
     def target(self):
         if self._target is None:
-            self._target = _parse_target(self.uncommented, self.dbapi2_module,
-                    self.operation)
+            self._target = _parse_target(self.uncommented, self.operation)
         return self._target
 
     @property
     def uncommented(self):
         if self._uncommented is None:
-            self._uncommented = _uncomment_sql(self.sql, self.dbapi2_module)
+            self._uncommented = _uncomment_sql(self.sql)
         return self._uncommented
 
     @property
     def obfuscated(self):
         if self._obfuscated is None:
-            self._obfuscated = _obfuscate_sql(self.uncommented,
-                    self.dbapi2_module)
+            self._obfuscated = _obfuscate_sql(self.uncommented, self.database)
         return self._obfuscated
 
     @property
     def normalized(self):
         if self._normalized is None:
-            self._normalized = _normalize_sql(self.obfuscated,
-                    self.dbapi2_module)
+            self._normalized = _normalize_sql(self.obfuscated)
         return self._normalized
 
     @property
@@ -602,21 +755,16 @@ class SQLStatement(object):
             self._identifier = hash(self.normalized)
         return self._identifier
 
-    def formatted(self, format):
-        if format == 'off':
+    def formatted(self, sql_format):
+        if sql_format == 'off':
             return ''
 
-        elif format == 'raw':
+        elif sql_format == 'raw':
             return self.sql
 
         else:
             return self.obfuscated
 
-    def explain_plan(self, connect_params, cursor_params, execute_params,
-            format):
-        return _explain_plan(self.sql, self.dbapi2_module, connect_params,
-                cursor_params, execute_params, format)
-
 _sql_statements = weakref.WeakValueDictionary()
 
 def sql_statement(sql, dbapi2_module):
@@ -627,7 +775,9 @@ def sql_statement(sql, dbapi2_module):
     if result is not None:
         return result
 
-    result = SQLStatement(sql, dbapi2_module)
+    database = SQLDatabase(dbapi2_module)
+    result = SQLStatement(sql, database)
+
     _sql_statements[key] = result
 
     return result
